﻿//
//  *** SGLIsample_20210506.cpp ***
//
#include <iostream>
#include <opencv2/opencv.hpp>
#include <hdf5.h>

using namespace std;

#define	uint16	unsigned short
#define	float32	float

int main()
{
	bool rtn = true;

	printf(" ... process start\n");

	string filename = "X:\\GC1SG1_202104100117C04810_1BSG_VNRDQ_2002.h5";
	string imggrp = "/Image_data";
	string sdname = "Lt_VN05";
	string l1b_png = "..\\" + sdname + "_L1B.png";
	string map_png = "..\\" + sdname + "_MAP.png";
	float max_ref = 0.2f;	// max reflectance
	float pngratio = 0.1f;	// png resize ratio
	float32	degStep = 0.002f;	// 0.002deg grid
	printf("target hdf : %s\n", filename.c_str());
	printf("target sd  : %s %s\n", imggrp.c_str(), sdname.c_str());

	//  HDF/Group/SD open
	hid_t fid = H5Fopen(filename.c_str(), H5F_ACC_RDWR, H5P_DEFAULT);
	hid_t gid = H5Gopen(fid, imggrp.c_str(), H5P_DEFAULT);
	hid_t sdid = H5Dopen(gid, sdname.c_str(), H5P_DEFAULT);

	hsize_t	dims[10];
	int n_dims = H5Sget_simple_extent_dims(H5Dget_space(sdid), dims, NULL);

	long lines = (long)dims[0];
	long pixels = (long)dims[1];

	cv::Mat dn(lines, pixels, CV_16UC1);	// SD Mat array(uint16)
	herr_t status = H5Dread(sdid, H5Dget_type(sdid), H5S_ALL, H5S_ALL, H5P_DEFAULT, dn.ptr());

	//  Attribute
	float32 Offset, Slope, TOA;
	uint16	Mask, MaxDN, MinDN;

	hid_t aid = H5Aopen_name(sdid, "Offset");
	status = H5Aread(aid, H5Aget_type(aid), &Offset);
	status = H5Aclose(aid);

	aid = H5Aopen_name(sdid, "Slope");
	status = H5Aread(aid, H5Aget_type(aid), &Slope);
	status = H5Aclose(aid);

	aid = H5Aopen_name(sdid, "Band_weighted_TOA_solar_irradiance");
	status = H5Aread(aid, H5Aget_type(aid), &TOA);
	status = H5Aclose(aid);

	aid = H5Aopen_name(sdid, "Mask");
	status = H5Aread(aid, H5Aget_type(aid), &Mask);
	status = H5Aclose(aid);

	aid = H5Aopen_name(sdid, "Maximum_valid_DN");
	status = H5Aread(aid, H5Aget_type(aid), &MaxDN);
	status = H5Aclose(aid);

	aid = H5Aopen_name(sdid, "Minimum_valid_DN");
	status = H5Aread(aid, H5Aget_type(aid), &MinDN);
	status = H5Aclose(aid);

	status = H5Dclose(sdid);
	status = H5Gclose(gid);

	//  lat, lon
	gid = H5Gopen(fid, "/Geometry_data", H5P_DEFAULT);
	sdid = H5Dopen(gid, "Latitude", H5P_DEFAULT);

	n_dims = H5Sget_simple_extent_dims(H5Dget_space(sdid), dims, NULL);
	long latlon_lines = (long)dims[0];	// resampled data
	long latlon_pixels = (long)dims[1];
	long respl = 10;					// resample intervarl = 10

	cv::Mat lat(latlon_lines, latlon_pixels, CV_32FC1);
	status = H5Dread(sdid, H5Dget_type(sdid), H5S_ALL, H5S_ALL, H5P_DEFAULT, lat.ptr());

	sdid = H5Dopen(gid, "Longitude", H5P_DEFAULT);
	cv::Mat lon(latlon_lines, latlon_pixels, CV_32FC1);
	status = H5Dread(sdid, H5Dget_type(sdid), H5S_ALL, H5S_ALL, H5P_DEFAULT, lon.ptr());

	//  HDF/Group/SD close
	status = H5Dclose(sdid);
	status = H5Gclose(gid);
	status = H5Fclose(fid);

	printf("pixels=%ld, lines=%ld\n", pixels, lines);
	printf("lat, lon : pixels=%ld, lines=%ld\n", latlon_pixels, latlon_lines);
	printf("attribute : Slope=%f/Offset=%f/Mask=%04X/MinDN=%d/MaxDN=%d\n", 
		Slope, Offset, Mask, MinDN, MaxDN);

	//  Engineering conversion
	constexpr float32 q_nan = std::numeric_limits<float>::quiet_NaN();
	cv::Mat img(cv::Size(pixels, lines), CV_32F, q_nan);	// image Mat array(float32)
	float32* ptr = (float32*)img.ptr();
	uint16* uptr = (uint16*)dn.ptr();

	for (long idx = 0; idx < lines * pixels; idx++) {
		uint16 d = uptr[idx] & Mask;
		ptr[idx] = (d >= MinDN && d <= MaxDN) ? ((float32)d * Slope + Offset) / TOA : q_nan;
	}
	dn.release();

	//  mat display
	cv::Mat dsp = img.clone() * 255.0 / max_ref;	// value range
	dsp.convertTo(dsp, CV_8U);
	applyColorMap(dsp, dsp, cv::COLORMAP_JET);
	cv::resize(dsp, dsp, cv::Size(), pngratio, pngratio);
	imwrite(l1b_png.c_str(), dsp);
	cv::imshow(sdname.c_str(), dsp);
	cv::waitKey(0);
	dsp.release();

	//  lat, lon range -> proj. Mat
	double minLat, maxLat, minLon, maxLon;
	cv::minMaxLoc(lon, &minLon, &maxLon);
	cv::minMaxLoc(lat, &minLat, &maxLat);
	minLat = (double) (long)(minLat / degStep + 0.5f) * degStep;
	maxLat = (double) (long)(maxLat / degStep + 0.5f) * degStep;
	minLon = (double) (long)(minLon / degStep + 0.5f) * degStep;
	maxLon = (double) (long)(maxLon / degStep + 0.5f) * degStep;

	long dst_pixs = (long)((maxLon - minLon) / degStep) + 1;
	long dst_lins = (long)((maxLat - minLat) / degStep) + 1;

	printf("proj step : %f [pix/deg]\n", degStep);
	printf("proj lon range : (%f:%f) -> %ld pixs\n", minLon, maxLon, dst_pixs);
	printf("proj lat range : (%f:%f) -> %ld lins\n", minLat, maxLat, dst_lins);

	//  proj array pointer
	cv::Mat dst(cv::Size(dst_pixs, dst_lins), CV_32FC1, q_nan);
	float32* dstPtr = (float32*)dst.ptr();
	long dstStep = (long)dst.step / sizeof(float32);

	float32* lonPtr = (float32*)lon.ptr();
	float32* latPtr = (float32*)lat.ptr();
	long latlonStep = (long)lon.step / sizeof(float32);

	//  map projection loop
	printf("... start of projection\n");
	for (float lin = 0; lin + (float)respl <= (float)lines - 3; lin += (float)respl) {
		for (float pix = 0; pix + (float)respl <= (float)pixels - 3; pix += (float)respl) {

			//  Source Array position
			cv::Point2f srcPnt[] =
			{
				cv::Point2f(pix,         lin),			// UL
				cv::Point2f(pix,         lin + respl),	// LL
				cv::Point2f(pix + respl, lin + respl),	// LR
				cv::Point2f(pix + respl, lin)			// UR
			};
			cv::Rect srcRect = cv::Rect(srcPnt[0], cv::Point2f(pix+respl+2, lin+respl+2)); // to prevent underlap
			cv::Mat srcMat = img(srcRect).clone();

			//  Destination Array position
			long idxUL = (long)(lin / respl) * latlonStep + (long)(pix / respl);
			long idxLL = (long)(lin / respl + 1) * latlonStep + (long)(pix / respl);
			long idxLR = (long)(lin / respl + 1) * latlonStep + (long)(pix / respl + 1);
			long idxUR = (long)(lin / respl) * latlonStep + (long)(pix / respl + 1);

			float32 lonUL = (lonPtr[idxUL] - (float)minLon) / (float)degStep;
			float32 lonLL = (lonPtr[idxLL] - (float)minLon) / (float)degStep;
			float32 lonLR = (lonPtr[idxLR] - (float)minLon) / (float)degStep;
			float32 lonUR = (lonPtr[idxUR] - (float)minLon) / (float)degStep;

			float32 latUL = ((float)maxLat - latPtr[idxUL]) / (float)degStep;
			float32 latLL = ((float)maxLat - latPtr[idxLL]) / (float)degStep;
			float32 latLR = ((float)maxLat - latPtr[idxLR]) / (float)degStep;
			float32 latUR = ((float)maxLat - latPtr[idxUR]) / (float)degStep;

			cv::Point2f dstPnt[] =
			{
				cv::Point2f(lonUL, latUL),
				cv::Point2f(lonLL, latLL),
				cv::Point2f(lonLR, latLR),
				cv::Point2f(lonUR, latUR)
			};
			float32 maxLonTmp = max(max(lonUL, lonLL), max(lonLR, lonUR));
			float32 minLonTmp = min(min(lonUL, lonLL), min(lonLR, lonUR));
			float32 maxLatTmp = max(max(latUL, latLL), max(latLR, latUR));
			float32 minLatTmp = min(min(latUL, latLL), min(latLR, latUR));
			long tmp_pixs = (long)(maxLonTmp - minLonTmp) + 2;	// to prevent under lap
			long tmp_lins = (long)(maxLatTmp - minLatTmp) + 2;

			//  matrix for warp calc
			cv::Point2f srcBase = srcPnt[0] = srcPnt[0] - cv::Point2f(0.5F, 0.5F);	// POA is necessary for UL
			cv::Point2f srcPntH[] =
			{
				srcPnt[0] - srcBase,	// UL
				srcPnt[1] - srcBase,	// LL
				srcPnt[2] - srcBase,	// LR
				srcPnt[3] - srcBase		// UR
			};
			cv::Point2f dstPntH[] =
			{
				dstPnt[0] - cv::Point2f(minLonTmp, minLatTmp),	// UL
				dstPnt[1] - cv::Point2f(minLonTmp, minLatTmp),	// LL
				dstPnt[2] - cv::Point2f(minLonTmp, minLatTmp),	// LR
				dstPnt[3] - cv::Point2f(minLonTmp, minLatTmp)	// UR
			};
			cv::Mat H = cv::getPerspectiveTransform(srcPntH, dstPntH);

			//  warp!
			cv::Mat tmpMat(cv::Size(tmp_pixs, tmp_lins), CV_32FC1, q_nan);
			float32* tmpPtr = (float32*)tmpMat.ptr();
			long tmpStep = (long)tmpMat.step / sizeof(float32);

			cv::warpPerspective(
				srcMat,
				tmpMat,
				H,
				tmpMat.size(),
				cv::INTER_LINEAR,
				cv::BORDER_TRANSPARENT);

			//  data merge
			for (long tmp_lin = 0; tmp_lin < tmp_lins; tmp_lin++)
				for (long tmp_pix = 0; tmp_pix < tmp_pixs; tmp_pix++) {
					long idx = (tmp_lin + (long) minLatTmp) * dstStep
						+ tmp_pix + (long)minLonTmp;
					long tmp_idx = tmp_lin * tmpStep + (long)tmp_pix;
					if (!std::isnan(tmpPtr[tmp_idx]))
						dstPtr[idx] = tmpPtr[tmp_idx];
				}
			tmpMat.release();
		}
	}

	//  map display
	dsp = dst.clone() * 255.0 / max_ref;	// max reflectance
	dsp.convertTo(dsp, CV_8U);
	applyColorMap(dsp, dsp, cv::COLORMAP_JET);
	cv::resize(dsp, dsp, cv::Size(), pngratio, pngratio);
	imwrite(map_png.c_str(), dsp);
	cv::imshow(sdname.c_str(), dsp);
	cv::waitKey(0);
	dsp.release();

	//  実行終了
	printf(" ... end of process\n\n");
	return true;
}
